# SPDX-License-Identifier: BSD-3-Clause
""" Common building blocks for Xilinx 7-series targets. """

from math import ceil
from amaranth import *

from .lfps import _LFPS_PERIOD_MAX


class DRPInterface:
    """ Dynamic Reconfiguration Port interface for Xilinx FPGAs. """

    def __init__(self):
        self.lock = Signal(1,   name="drp_lock")
        self.addr = Signal(9,   name="drp_addr")
        self.di   = Signal(16,  name="drp_di")
        self.do   = Signal(16,  name="drp_do")
        self.we   = Signal(1,   name="drp_we")
        self.en   = Signal(1,   name="drp_en")
        self.rdy  = Signal(1,   name="drp_rdy")


class _DRPInterfaceBuffer(Elaboratable):
    """ Gateware that latches DRP transaction inputs, for an arbiter to complete it later. """
    def __init__(self, interface):
        self.intf = interface

        self.addr_latch = Signal.like(interface.addr)
        self.di_latch = Signal.like(interface.di)
        self.we_latch = Signal.like(interface.we)
        self.en_latch = Signal.like(interface.en)


    def elaborate(self, platform):
        m = Module()

        with m.If(self.intf.en):
            m.d.ss += [
                self.addr_latch.eq(self.intf.addr),
                self.di_latch.eq(self.intf.di),
                self.we_latch.eq(self.intf.we),
                self.en_latch.eq(1),
            ]

        with m.Elif(self.intf.rdy):
            m.d.ss += [
                self.en_latch.eq(0),
            ]

        return m


class DRPArbiter(Elaboratable):
    """ Gateware that merges a collection of DRPInterfaces into a single interface.

    To support safe read-modify-write operations, the ``lock`` signal can be used to gain
    exclusive access to the reconfiguration port. After starting a DRP operation with ``lock``
    asserted, and until it is deasserted, no other client will be able to access the port.
    """

    def __init__(self):
        self.shared = DRPInterface()
        self.interfaces = []

    def add_interface(self, interface: DRPInterface):
        self.interfaces.append(interface)


    def elaborate(self, platform):
        m = Module()

        buffers = Array(_DRPInterfaceBuffer(intf) for intf in self.interfaces)
        m.submodules += buffers

        current_idx = Signal(range(len(buffers)))
        current_buf = buffers[current_idx]

        with m.FSM(domain="ss"):

            with m.State("IDLE"):
                for idx in range(len(buffers)):
                    with m.If(buffers[idx].en_latch):
                        m.d.ss += current_idx.eq(idx)
                        m.next = "REQUEST"

            with m.State("REQUEST"):
                m.d.comb += [
                    self.shared.lock.eq(current_buf.intf.lock),
                    self.shared.addr.eq(current_buf.addr_latch),
                    self.shared.di.eq(current_buf.di_latch),
                    self.shared.we.eq(current_buf.we_latch),
                    self.shared.en.eq(current_buf.en_latch),
                ]
                with m.If(self.shared.en):
                    m.next = "REPLY"
                with m.Elif(~current_buf.intf.lock):
                    m.next = "IDLE"

            with m.State("REPLY"):
                m.d.comb += [
                    current_buf.intf.do.eq(self.shared.do),
                    current_buf.intf.rdy.eq(self.shared.rdy),
                ]
                with m.If(self.shared.rdy):
                    with m.If(current_buf.intf.lock):
                        m.next = "REQUEST"
                    with m.Else():
                        m.next = "IDLE"

        return m


class DRPFieldController(Elaboratable):
    """ Gateware that atomically updates part of a word via DRP. """

    def __init__(self, *, addr: int, bits: slice, init=0):
        self._addr = addr
        self._bits = bits

        self.drp = DRPInterface()

        self.value = Signal(bits.stop - bits.start, init=reset)


    def elaborate(self, platform):
        m = Module()

        m.d.comb += [
            self.drp.lock.eq(1),
            self.drp.addr.eq(self._addr)
        ]

        current_val = Signal.like(self.drp.do)

        with m.FSM(domain="ss"):
            with m.State("READ"):
                m.d.comb += [
                    self.drp.en.eq(1)
                ]
                m.next = "READ-WAIT"

            with m.State("READ-WAIT"):
                with m.If(self.drp.rdy):
                    m.d.ss += [
                        current_val.eq(self.drp.do),
                        current_val[self._bits].eq(self.value)
                    ]
                    m.next = "WRITE"

            with m.State("WRITE"):
                m.d.comb += [
                    self.drp.di.eq(current_val),
                    self.drp.we.eq(1),
                    self.drp.en.eq(1),
                ]
                m.next = "WRITE-WAIT"

            with m.State("WRITE-WAIT"):
                with m.If(self.drp.rdy):
                    m.next = "IDLE"

            with m.State("IDLE"):
                m.d.comb += self.drp.lock.eq(0)
                with m.If(current_val[self._bits] != self.value):
                    m.next = "READ"

        return m


class GTResetDeferrer(Elaboratable):
    """ Gateware that ensures the mandatory post-configuration period before reset, per Xilinx AR43482. """

    def __init__(self, ss_clock_frequency):
        self._ss_clock_frequency = ss_clock_frequency

        # Reset inputs.
        self.tx_i = Signal()
        self.rx_i = Signal()

        # Reset outputs.
        self.tx_o = Signal()
        self.rx_o = Signal()

        # Status output.
        self.done = Signal()


    def elaborate(self, platform):
        m = Module()

        # Defer reset by 500ns recommended in [AR43482], plus 5% margin.
        cycles = int(self._ss_clock_frequency * 500e-9 * 1.05)
        timer  = Signal(range(cycles))

        # Defer reset immediately after configuration; and never again, even if our domain is reset.
        defer  = Signal(init=1, reset_less=True)

        with m.If(defer):
            m.d.ss += timer.eq(timer + 1)

            with m.If(timer + 1 == cycles):
                m.d.ss += defer.eq(0)

        with m.Else():
            m.d.comb += [
                self.done.eq(1),
                self.tx_o.eq(self.tx_i),
                self.rx_o.eq(self.rx_i),
            ]

        return m


class GTPRXPMAResetWorkaround(Elaboratable):
    """ Gateware that ensures the required conditions for GTP receiver are met, per UG482. """

    def __init__(self, ss_clock_frequency):
        self._ss_clock_frequency = ss_clock_frequency

        self.i = Signal()
        self.o = Signal()

        self.rxpmaresetdone = Signal()
        self.drp = DRPInterface()


    def elaborate(self, platform):
        m = Module()

        m.d.comb += [
            self.drp.lock.eq(1),
            self.drp.addr.eq(0x011)
        ]

        saved_val = Signal.like(self.drp.do)

        with m.FSM(domain="ss"):
            with m.State("IDLE"):
                m.d.comb += [
                    self.drp.lock.eq(0),
                ]
                with m.If(self.i):
                    m.next = "READ"

            with m.State("READ"):
                m.d.comb += [
                    self.o.eq(1),
                    self.drp.en.eq(1)
                ]
                m.next = "READ-WAIT"

            with m.State("READ-WAIT"):
                m.d.comb += [
                    self.o.eq(1),
                ]
                with m.If(self.drp.rdy):
                    m.d.ss += [
                        saved_val.eq(self.drp.do)
                    ]
                    m.next = "WRITE"

            with m.State("WRITE"):
                m.d.comb += [
                    self.o.eq(1),
                ]
                m.d.comb += [
                    self.drp.di.eq(saved_val),
                    self.drp.di[11].eq(0),
                    self.drp.we.eq(1),
                    self.drp.en.eq(1),
                ]
                m.next = "WRITE-WAIT"

            with m.State("WRITE-WAIT"):
                m.d.comb += [
                    self.o.eq(1),
                ]
                with m.If(self.drp.rdy):
                    m.next = "RESET-WAIT"

            with m.State("RESET-WAIT"):
                m.d.comb += [
                    self.o.eq(1),
                ]
                with m.If(~self.i):
                    m.next = "RXPMARESETDONE-WAIT"

            with m.State("RXPMARESETDONE-WAIT"):
                with m.If(self.rxpmaresetdone):
                    m.next = "RESTORE"
                with m.If(self.i):
                    m.next = "READ"

            with m.State("RESTORE"):
                with m.If(~self.rxpmaresetdone):
                    m.d.comb += [
                        self.drp.di.eq(saved_val),
                        self.drp.we.eq(1),
                        self.drp.en.eq(1),
                    ]
                    m.next = "RESTORE-WAIT"
                with m.If(self.i):
                    m.next = "READ"

            with m.State("RESTORE-WAIT"):
                with m.If(self.drp.rdy):
                    m.next = "IDLE"
                with m.If(self.i):
                    m.next = "READ"

        return m


# Unlike e.g. Lattice ECP5 series, Xilinx's SerDes do not have a direct path from
# the high-speed pads to the fabric; which makes detecting out-of-band signaling
# challenging. As an (undocumented) alternative to the method described below, on
# GTX transceivers, when the RXQPIEN input is driven high, the RXQPISENP/RXQPISENN
# outputs are connected to (single-ended) input buffers for the corresponding pad,
# and can be used from the fabric. On GTP transceivers these ports also exist,
# under different names: PMARSVDOUT0/PMARSVDOUT1 and PMARSVDIN2, respectively.

class GTOOBClockDivider(Elaboratable):
    """ Gateware that derives a clock for the out-of-band detector suitable for demodulating LFPS signaling. """

    # Out-of-band decoding requires an auxiliary clock with a specific frequency.
    # [UG482] does not describe how the Series 7 GTX OOB detector functions; however,
    # [UG196] describes in detail how the Virtex-5 GTP OOB detector works, which closely
    # matches the observable behavior of the GTX block.
    #
    # The input of the OOB detector is first processed through a peak detector, and then
    # sampled on both edges of a user-provided clock. The link is considered idle at
    # any time when the three last sampling intervals contained no transitions.
    #
    # LFPS sequences consist of alternating square wave bursts and periods of electrical
    # idle. If the clock driving the OOB detector is slow enough that the longest allowed
    # LFPS half-period is longer than three sampling intervals, then an LFPS sequence will
    # be demodulated by the OOB detector; RXELECIDLE will be low during the burst, and
    # high during the remainder of the repeat period.

    def __init__(self, ss_clock_frequency, max_pulse_period=_LFPS_PERIOD_MAX):
        # The generated clock must sample each half-period less than three times;
        # so that no matter how the OOB detector clock aligns with the LFPS waveform,
        # there would always be a transition.
        self._ratio = ceil((ss_clock_frequency * 3) * (max_pulse_period / 2))

        self.o = Signal()


    def elaborate(self, platform):
        m = Module()

        counter = Signal(range(self._ratio))

        m.d.ss += counter.eq(counter + 1)
        with m.If(counter + 1 == self._ratio):
            m.d.ss += [
                counter.eq(0),
                self.o.eq(~self.o)
            ]

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
from luna.gateware.test.utils import LunaGatewareTestCase, sync_test_case

from amaranth import Signal
from luna.gateware.interface.spi import SPIDeviceInterface, SPIRegisterInterface, SPIGatewareTestCase

class SPIDeviceInterfaceTest(SPIGatewareTestCase):
    FRAGMENT_UNDER_TEST = SPIDeviceInterface
    FRAGMENT_ARGUMENTS = dict(word_size=16, clock_polarity=1)

    def initialize_signals(self):
        yield self.dut.spi.cs.eq(0)


    @sync_test_case
    def test_spi_interface(self):

        # Ensure that we don't complete a word while CS is deasserted.
        for _ in range(10):
            self.assertEqual((yield self.dut.word_complete), 0)
            yield

        # Set the word we're expected to send, and then assert CS.
        yield self.dut.word_out.eq(0xABCD)
        yield

        yield self.dut.spi.cs.eq(1)
        yield

        # Verify that the SPI in/out behavior is what we expect.
        response = yield from self.spi_exchange_data(b"\xCA\xFE")
        self.assertEqual(response, b"\xAB\xCD")
        self.assertEqual((yield self.dut.word_in), 0xCAFE)


    @sync_test_case
    def test_spi_transmit_second_word(self):

        # Set the word we're expected to send, and then assert CS.
        yield self.dut.word_out.eq(0x0f00)
        yield

        yield self.dut.spi.cs.eq(1)
        yield

        # Verify that the SPI in/out behavior is what we expect.
        response = yield from self.spi_exchange_data(b"\x00\x00")
        self.assertEqual(response, b"\x0F\x00")



class SPIRegisterInterfaceTest(SPIGatewareTestCase):
    """ Tests for the SPI command interface. """

    def instantiate_dut(self):

        self.write_strobe = Signal()

        # Create a register and sample dataset to work with.
        dut = SPIRegisterInterface(default_read_value=0xDEADBEEF)
        dut.add_register(2, write_strobe=self.write_strobe)

        return dut


    def initialize_signals(self):
        # Start off with our clock low and the transaction idle.
        yield self.dut.spi.sck.eq(0)
        yield self.dut.spi.cs.eq(0)


    @sync_test_case
    def test_undefined_read_behavior(self):
        data = yield from self.spi_exchange_data([0, 1, 0, 0, 0, 0])
        self.assertEqual(bytes(data), b"\x00\x00\xde\xad\xbe\xef")


    @sync_test_case
    def test_write_behavior(self):

        # Send a write command...
        data = yield from self.spi_exchange_data(b"\x80\x02\x12\x34\x56\x78")
        self.assertEqual(bytes(data), b"\x00\x00\x00\x00\x00\x00")

        # ... and then read the relevant data back.
        data = yield from self.spi_exchange_data(b"\x00\x02\x12\x34\x56\x78")
        self.assertEqual(bytes(data), b"\x00\x00\x12\x34\x56\x78")


    @sync_test_case
    def test_aborted_write_behavior(self):

        # Set an initial value...
        data = yield from self.spi_exchange_data(b"\x80\x02\x12\x34\x56\x78")

        # ... and then perform an incomplete write.
        data = yield from self.spi_exchange_data(b"\x80\x02\xAA\xBB")

        # We should return to being idle after CS is de-asserted...
        yield
        self.assertEqual((yield self.dut.idle), 1)

        # ... and our register data should not have changed.
        data = yield from self.spi_exchange_data(b"\x00\x02\x12\x34\x56\x78")
        self.assertEqual(bytes(data), b"\x00\x00\x12\x34\x56\x78")



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Gateware for creating USB2 devices. """

from enum import IntEnum, IntFlag

#
# TODO: these should be moved into a USBPacketID library
#

class USBSpeed(IntEnum):
    """ Enumeration representing USB speeds. Matches UTMI xcvr_select constants. """

    HIGH = 0b00
    FULL = 0b01
    LOW  = 0b10



class USBPacketID(IntFlag):
    """ Enumeration specifying all of the valid USB PIDs we can handle. """

    # Token group (lsbs = 0b01).
    OUT   = 0b0001
    IN    = 0b1001
    SOF   = 0b0101
    SETUP = 0b1101

    # Data group (lsbs = 0b11).
    DATA0 = 0b0011
    DATA1 = 0b1011
    DATA2 = 0b0111
    MDATA = 0b1111

    # Handshake group (lsbs = 0b10)
    ACK   = 0b0010
    NAK   = 0b1010
    STALL = 0b1110
    NYET  = 0b0110

    # Special group.
    PRE   = 0b1100
    ERR   = 0b1100
    SPLIT = 0b1000
    PING  = 0b0100

    # Flag representing that the PID seems invalid.
    PID_INVALID   = 0b10000
    PID_CORE_MASK = 0b01111


    @classmethod
    def from_byte(cls, byte, skip_checks=False):
        """ Creates a PID object from a byte. """

        # Convert the raw PID to an integer.
        pid_as_int = int.from_bytes(byte, byteorder='little')
        return cls.from_int(pid_as_int, skip_checks=skip_checks)


    @classmethod
    def from_int(cls, value, skip_checks=True):
        """ Create a PID object from an integer. """

        PID_MASK           = 0b1111
        INVERTED_PID_SHIFT = 4

        # Pull out the PID and its inverse from the byte.
        pid          = cls(value & PID_MASK)
        inverted_pid = value >> INVERTED_PID_SHIFT

        # If we're not skipping checks,
        if not skip_checks:
            if (pid ^ inverted_pid) != PID_MASK:
                pid |= cls.PID_INVALID

        return cls(pid)


    @classmethod
    def from_name(cls, name):
        """ Create a PID object from a string representation of its name. """
        return cls[name]


    @classmethod
    def parse(cls, value):
        """ Attempt to create a PID object from a number, byte, or string. """

        if isinstance(value, bytes):
            return cls.from_byte(value)

        if isinstance(value, str):
            return cls.from_name(value)

        if isinstance(value, int):
            return cls.from_int(value)

        return cls(value)


    def category(self):
        """ Returns the USBPIDCategory that each given PID belongs to. """
        return USBPIDCategory(self & USBPIDCategory.MASK)


    def is_data(self):
        """ Returns true iff the given PID represents a DATA packet. """
        return self.category() is USBPIDCategory.DATA


    def is_token(self):
        """ Returns true iff the given PID represents a token packet. """
        return self.category() is USBPIDCategory.TOKEN


    def is_handshake(self):
        """ Returns true iff the given PID represents a handshake packet. """
        return self.category() is USBPIDCategory.HANDSHAKE


    def is_invalid(self):
        """ Returns true if this object is an attempt to encapsulate an invalid PID. """
        return (self & self.PID_INVALID)

    def direction(self):
        """ Get a USB direction from a PacketID. """

        if self is self.SOF:
            return None

        if self is self.SETUP or self is self.OUT:
            return USBDirection.OUT

        if self is self.IN:
            return USBDirection.IN

        raise ValueError("cannot determine the direction of a non-token PID")



    def summarize(self):
        """ Return a summary of the given packet. """

        # By default, get the raw name.
        core_pid  = self & self.PID_CORE_MASK
        name = core_pid.name

        if self.is_invalid():
            return "{} (check-nibble invalid)".format(name)
        else:
            return name


    def byte(self):
        """ Return the value with its upper nibble. """

        inverted_pid = int(self) ^ 0b1111
        full_pid     = (inverted_pid << 4) | int(self)

        return full_pid



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" USB3 link-layer abstraction."""

from amaranth import *

from ..link.header    import HeaderQueueArbiter, HeaderQueueDemultiplexer
from ...stream        import USBRawSuperSpeedStream

from .link_management import LinkManagementPacketHandler
from .timestamp       import TimestampPacketReceiver
from .endpoint        import SuperSpeedEndpointInterface
from .transaction     import TransactionPacketGenerator, TransactionPacketReceiver
from .data            import DataHeaderReceiver

class USB3ProtocolLayer(Elaboratable):
    """ Abstraction encapsulating the USB3 protocol layer hardware. """

    def __init__(self, *, link_layer):
        self._link = link_layer

        #
        # I/O port
        #

        # Interface to our various endpoints.
        self.endpoint_interface    = SuperSpeedEndpointInterface()

        # Device state inputs.
        self.current_address       = Signal(7)
        self.current_configuration = Signal(7)

        # Current timestamp.
        self.bus_interval          = Signal(14)


    def elaborate(self, platform):
        m = Module()
        link = self._link

        #
        # Header Packet Multiplexers
        #

        # One-to-many Header demultiplexer.
        m.submodules.hp_demux = hp_demux = HeaderQueueDemultiplexer()
        m.d.comb += hp_demux.sink.header_eq(link.header_source)

        # Many-to-one Header multiplexer.
        m.submodules.hp_mux = hp_mux = HeaderQueueArbiter()
        m.d.comb += link.header_sink.header_eq(hp_mux.source)


        #
        # Link Management Packet Handler
        #
        m.submodules.lmp_handler = lmp_handler = LinkManagementPacketHandler()

        hp_demux.add_consumer(lmp_handler.header_sink)
        hp_mux.add_producer(lmp_handler.header_source)

        m.d.comb += [
            lmp_handler.usb_reset     .eq(link.in_reset),
            lmp_handler.link_ready    .eq(link.ready),
        ]

        #
        # Isochronous Timestamp Packet Handler
        #
        m.submodules.itp_handler = itp_handler = TimestampPacketReceiver()
        hp_demux.add_consumer(itp_handler.header_sink)

        m.d.comb += [
            self.bus_interval.eq(itp_handler.bus_interval_counter)
        ]


        #
        # Data Packet Handlers
        #
        m.submodules.data_header_receiver = data_header_receiver = DataHeaderReceiver()
        hp_demux.add_consumer(data_header_receiver.header_sink)


        #
        # Transaction Packet handlers.
        #

        # Generator
        m.submodules.tp_generator = tp_generator = TransactionPacketGenerator()
        hp_mux.add_producer(tp_generator.header_source)
        m.d.comb += [
            tp_generator.address.eq(self.current_address)
        ]

        # Receiver
        m.submodules.tp_receiver = tp_receiver = TransactionPacketReceiver()
        hp_demux.add_consumer(tp_receiver.header_sink)



        #
        # Endpoint Interfacing
        #
        endpoint_interface = self.endpoint_interface
        m.d.comb += [
            # Rx interface.
            endpoint_interface.rx           .tap(link.data_source),
            endpoint_interface.rx_header    .eq(link.data_header_from_host),
            endpoint_interface.rx_complete  .eq(link.data_source_complete),
            endpoint_interface.rx_invalid   .eq(link.data_source_invalid),

            # Tx interface.
            link.data_sink                  .stream_eq(endpoint_interface.tx),
            link.data_sink_send_zlp         .eq(endpoint_interface.tx_zlp),
            link.data_sink_length           .eq(endpoint_interface.tx_length),
            link.data_sink_endpoint_number  .eq(endpoint_interface.tx_endpoint_number),
            link.data_sink_sequence_number  .eq(endpoint_interface.tx_sequence_number),
            link.data_sink_direction        .eq(endpoint_interface.tx_direction),

            # Handshake exchange interface.
            tp_generator.interface          .connect(endpoint_interface.handshakes_out),
            tp_receiver.interface           .connect(endpoint_interface.handshakes_in)
        ]


        return m



#
# This file is part of LUNA.
#

import os
import sys
import shutil
import logging
import tempfile
import argparse

from amaranth           import Elaboratable
from amaranth._unused   import MustUse

from luna.gateware.platform import configure_toolchain

# Log formatting strings.
LOG_FORMAT_COLOR = "\u001b[37;1m%(levelname)-8s| \u001b[0m\u001b[1m%(module)-12s|\u001b[0m %(message)s"
LOG_FORMAT_PLAIN = "%(levelname)-8s:n%(module)-12s>%(message)s"


def configure_default_logging(level=logging.INFO, logger=logging):

    # Set up our logging / output.
    if sys.stdout.isatty():
        log_format = LOG_FORMAT_COLOR
    else:
        log_format = LOG_FORMAT_PLAIN

    logger.basicConfig(level=level, format=log_format)


def top_level_cli(fragment, *pos_args, **kwargs):
    from .gateware.platform import get_appropriate_platform

    """ Runs a default CLI that assists in building and running gateware.

        If the user's options resulted in the board being programmed, this returns the fragment
        that was programmed onto the board. Otherwise, it returns None.

        Parameters:
            fragment  -- The fragment instance to be built; or a callable that returns a fragment,
                         such as a Elaborable type. If the latter is provided, any keyword or positional
                         arguments not specified here will be passed to this callable.
    """

    name = fragment.__name__ if callable(fragment) else fragment.__class__.__name__

    parser = argparse.ArgumentParser(description=f"Gateware generation/upload script for '{name}' gateware.")
    parser.add_argument('--output', '-o', metavar='filename', help="Build and output a bitstream to the given file.")
    parser.add_argument('--erase', '-E', action='store_true',
         help="Clears the relevant FPGA's flash before performing other options.")
    parser.add_argument('--upload', '-U', action='store_true',
         help="Uploads the relevant design to the target hardware. Default if no options are provided.")
    parser.add_argument('--flash', '-F', action='store_true',
         help="Flashes the relevant design to the target hardware's configuration flash.")
    parser.add_argument('--dry-run', '-D', action='store_true',
         help="When provided as the only option; builds the relevant bitstream without uploading or flashing it.")
    parser.add_argument('--keep-files', action='store_true',
         help="Keeps the local files in the default `build` folder.")
    parser.add_argument('--fpga', metavar='part_number',
         help="Overrides build configuration to build for a given FPGA. Useful if no FPGA is connected during build.")
    parser.add_argument('--console', metavar="port",
         help="Attempts to open a convenience 115200 8N1 UART console on the specified port immediately after uploading.")

    # Disable UnusedElaboarable warnings until we decide to build things.
    # This is sort of cursed, but it keeps us categorically from getting UnusedElaborable warnings
    # if we're not actually buliding.
    MustUse._MustUse__silence = True

    args = parser.parse_args()
    configure_default_logging()

    # If this isn't a fragment directly, interpret it as an object that will build one.
    if callable(fragment):
        fragment = fragment(*pos_args, **kwargs)

    # If we have no other options set, build and upload the relevant file.
    if (args.output is None and not args.flash and not args.erase and not args.dry_run):
        args.upload = True

    # Once the device is flashed, it will self-reconfigure, so we
    # don't need an explicitly upload step; and it implicitly erases
    # the flash, so we don't need an erase step.
    if args.flash:
        args.erase = False
        args.upload = False


    # Build the relevant gateware, uploading if requested.
    build_dir = "build" if args.keep_files else tempfile.mkdtemp()

    # Build the relevant files.
    try:
        platform = get_appropriate_platform()

        # If we have a toolchain override, apply it to our platform.
        toolchain = os.getenv("LUNA_TOOLCHAIN")
        if toolchain:
            platform.toolchain = toolchain

        if args.fpga:
            platform.device = args.fpga

        if args.erase:
            logging.info("Erasing flash...")
            platform.toolchain_erase()
            logging.info("Erase complete.")

        join_text = "and uploading gateware to attached" if args.upload else "for"
        logging.info(f"Building {join_text} {platform.name}...")

        # Configure toolchain.
        if not configure_toolchain(platform):
            logging.info(f"Failed to configure the toolchain for: {platform.toolchain}")
            logging.info(f"Continuing anyway.")

        # Now that we're actually building, re-enable Unused warnings.
        MustUse._MustUse__silence = False

        # Perform the build.
        products = platform.build(fragment,
            do_program=args.upload,
            build_dir=build_dir
        )

        logging.info(f"{'Upload' if args.upload else 'Build'} complete.")

        # If we're flashing the FPGA's flash, do so.
        if args.flash:
            logging.info("Programming flash...")
            platform.toolchain_flash(products)
            logging.info("Programming complete.")

        # If we're outputting a file, write it.
        if args.output:
            bitstream =  products.get("top.bit")
            with open(args.output, "wb") as f:
                f.write(bitstream)

        # If we're expecting a console, open one.
        if args.console:
            import serial.tools.miniterm

            # Clear our arguments, so they're not parsed by miniterm.
            del sys.argv[1:]

            # Run miniterm with our default port and baudrate.
            serial.tools.miniterm.main(default_port=args.console, default_baudrate=115200)

        # Return the fragment we're working with, for convenience.
        if args.upload or args.flash:
            return fragment

    # Clean up any directories we've created.
    finally:
        if not args.keep_files:
            shutil.rmtree(build_dir)

    return None



#
# This file is part of LUNA.
#
# Copyright (c) 2024 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

import struct

from amaranth                                     	import DomainRenamer, Elaboratable, Module, Signal
from amaranth.lib.memory                            import Memory
from usb_protocol.emitters.descriptors.microsoft10 	import MicrosoftOS10DescriptorCollection

from ...stream                                    	import USBInStreamInterface


class GetMicrosoftDescriptorHandlerBlock(Elaboratable):
    """ Gateware that handles responding to GET_MS_DESCRIPTOR requests.

    I/O port:
        I: request[8]     -- The index field associated with the GET_MS_DESCRIPTOR request.
        I: length[16]     -- The length field associated with the GET_MS_DESCRIPTOR request.
                             Determines the maximum amount allowed in a response.

        I: start          -- Strobe that indicates when a descriptor should be transmitted.
        I: start_position -- Specifies the starting position of the descriptor data to be transmitted.

        *: tx             -- The USBInStreamInterface that streams our descriptor data.
        O: stall          -- Pulsed if a STALL handshake should be generated, instead of a response.
    """

    ELEMENT_SIZE = 4

    COUNT_SIZE_BITS   = 16
    ADDRESS_SIZE_BITS = 16

    def __init__(self, descriptor_collection: MicrosoftOS10DescriptorCollection, max_packet_length=64, domain="usb"):
        """
        Parameters
        ----------
        descriptor_collection: MicrosoftOS10DescriptorCollection
            The MicrosoftOS10DescriptorCollection containing the descriptors to use for this device.
        max_packet_length: int
            Maximum packet length.
        domain: string
            The clock domain this generator should belong to. Defaults to 'usb'.
        """

        self._descriptors        = descriptor_collection
        self._max_packet_length  = max_packet_length
        self._domain             = domain

        #
        # I/O port
        #
        self.index          = Signal(8)
        self.length         = Signal(16)

        self.start          = Signal()
        self.start_position = Signal(11)

        self.tx             = USBInStreamInterface()
        self.stall          = Signal()


    @classmethod
    def _align_to_element_size(cls, n):
        """ Returns a given number rounded up to the next "aligned" element size. """
        return (n + (cls.ELEMENT_SIZE - 1)) // cls.ELEMENT_SIZE

    def generate_rom_content(self):
        """ Generates the contents of the ROM used to hold descriptors.

        Memory layout
        -------------

        All data is aligned on 4 bytes

        Index offsets and length of descriptor
        --------------------------------------
        Each index of a descriptor type has an entry consistent of the length
        of the descriptor (2 bytes) and the address of first data byte (2 bytes).

        0010  Length of first device descriptor
        0012  Address of first device descriptor
        ...

        Data
        ----
        Descriptor data for each descriptor. Padded by 0 to next 4-byte address.

        ...   Descriptor data

        """

        # Get all descriptors and cache them in a dictionary, so that we can access them at will.
        descriptors = {}
        for index, raw_descriptor in self._descriptors:
            descriptors[index] = raw_descriptor

        # For now, we only support layouts with consecutive indexes.
        # Ensure this is the case.
        assert max(descriptors.keys()) - min(descriptors.keys()) == len(descriptors) - 1, \
            "descriptors have non-contiguous indices!"

        #
        # Compute the ROM size that we'll need.
        #
        max_index_number        = max(descriptors.keys())
        min_index_number        = min(descriptors.keys())
        indexes_count           = max_index_number - min_index_number + 1
        max_descriptor_size     = 0

        # Our ROM starts with a collection of pointers to our various descriptor tables...
        rom_size_table_pointers = indexes_count * self.ELEMENT_SIZE

        # ... and the descriptors themselves.
        rom_size_descriptors = 0
        for raw_descriptor in descriptors.values():

            # Compute the maximum size for each descriptor...
            aligned_size = self._align_to_element_size(len(raw_descriptor))
            rom_size_descriptors += aligned_size * self.ELEMENT_SIZE

            # ... and store the maximum size we've encountered.
            max_descriptor_size = max(max_descriptor_size, len(raw_descriptor))

        # Create an array to hold our initial values for our composite ROM.
        total_size = \
            rom_size_table_pointers + \
            rom_size_descriptors
        rom = bytearray(total_size)

        #
        # Fill the ROM's initialization values.
        #
        next_free_address       = rom_size_table_pointers

        # Next, create the tables themselves, which are filled with data pointers,
        # and add our descriptors to our memory.
        for index, raw_descriptor in sorted(descriptors.items()):

            # Create our descriptor pointer entries...
            pointer_bytes = struct.pack(">HH", len(raw_descriptor), next_free_address)

            # ... figure out where in the ROM we're going to store the pointer ...
            index_base_address = (index - min_index_number) * self.ELEMENT_SIZE

            # ... add the pointer...
            rom[index_base_address:index_base_address + 4] = pointer_bytes

            # ... and then store the descriptor itself to the pointer address.
            rom[next_free_address:next_free_address+len(raw_descriptor)] = raw_descriptor

            # Figure out the next free position for a descriptor.
            aligned_size = self._align_to_element_size(len(raw_descriptor))
            next_free_address += aligned_size * self.ELEMENT_SIZE

        assert total_size == len(rom)


        #
        # Finally, convert our ROM into an initialization vector.
        #
        total_elements = total_size // self.ELEMENT_SIZE
        element_size = self.ELEMENT_SIZE

        # Chunk our ROM into a collection of entries...
        rom_entries = (rom[(element_size * i):(element_size * i) + element_size] for i in range(total_elements))

        # ... and then convert that into an initializer value in the format Amaranth ROMs like (integers).
        initializer = [struct.unpack(">I", rom_entry)[0] for rom_entry in rom_entries]

        return initializer, max_descriptor_size, max_index_number, min_index_number


    def elaborate(self, platform):
        m = Module()

        index = self.index

        #
        # Create the ROM that stores our descriptors...
        #
        rom_content, descriptor_max_length, max_index, min_index = self.generate_rom_content()

        m.submodules.rom = rom = Memory(shape=32, depth=len(rom_content), init=rom_content)
        rom_read_port = rom.read_port()

        # Create convenience aliases to the upper and lower half of the ROM.
        rom_upper_half = rom_read_port.data.word_select(1, 16)
        rom_lower_half = rom_read_port.data.word_select(0, 16)

        # All of our ROM's metadata is composed of elements formatted as (count, pointer).
        # Grab a quick reference to the ROM's upper half, which stores the count...
        rom_element_count    = rom_upper_half

        # ... and to the ROM's lower half, not counting the last two bits (which are always 0,
        # as our pointers are always aligned). This creates an element pointer counted in words,
        # instead of in bytes; and thus one compatible with our read_port addr.
        rom_element_pointer  = rom_read_port.data.bit_select(2, rom_read_port.addr.width)

        #
        # Figure out the maximum length we're willing to send.
        #
        length = Signal(16)

        # We'll never send more than our MaxPacketSize. This means that we'll want to send a maximum of
        # either our maximum packet length, or the amount of data we have remaining; whichever is less.
        #
        words_remaining = self.length - self.start_position
        with m.If(words_remaining <= self._max_packet_length):
            m.d.sync += length.eq(words_remaining)
        with m.Else():
            m.d.sync += length.eq(self._max_packet_length)

        # Register that stores our current position in the stream.
        # We still want to be able to store a position beyond bounds (+1),
        # this is required for descriptors length multiple of the maximum packet size.
        # Like this we do not overflow our position and are able to send a ZLP on the next request.
        position_in_stream = Signal(range(descriptor_max_length + 1))
        bytes_sent = Signal.like(length)

        # Registers that store descriptor length and data base address.
        descriptor_length = Signal(16)
        descriptor_data_base_address = Signal(rom_read_port.addr.width)

        # Track when we're on the first and last packet.
        on_first_packet = position_in_stream == self.start_position
        on_last_packet = \
            (position_in_stream == (descriptor_length - 1)) | \
            (bytes_sent + 1 >= length)

        #
        # Core transmit logic.
        #

        with m.FSM():

            # IDLE -- we're currently waiting to send a descriptor.
            with m.State('IDLE'):

                # Reset our data-sent count...
                m.d.sync += bytes_sent.eq(0)

                # ... and always prepare to read whatever descriptor type is requested.
                m.d.comb += rom_read_port.addr.eq(index - min_index)

                # Once we have a request to start transmitting...
                with m.If(self.start):
                    m.next = 'START'

            # START -- retiming state to allow construction of the length signal
            with m.State('START'):
                # ... and always prepare to read whatever descriptor type is requested.
                m.d.comb += rom_read_port.addr.eq(index - min_index)

                # ... apply our start position...
                m.d.sync += position_in_stream.eq(self.start_position),

                is_valid_index = (min_index <= index) & (index <= max_index)

                # If we have a descriptor we're able to send, prepare to send it.
                with m.If(is_valid_index):
                    m.next = 'LOOKUP_DESCRIPTOR'

                # Otherwise, stall the request immediately.
                with m.Else():
                    m.d.comb += self.stall.eq(1)
                    m.next = 'IDLE'


            # LOOKUP_DESCRIPTOR -- we've now fetched from ROM the location of the descriptor in memory.
            # We'll decode it, and then prepare to start sending the descriptor.
            # descriptor from memory. First, we'll need to find the location of the table that contains each
            # descriptor pointer.
            with m.State('LOOKUP_DESCRIPTOR'):

                # Point our descriptor at the first word in our descriptor, offset by our current position
                # in the stream...
                m.d.comb += rom_read_port.addr.eq((rom_read_port.data + position_in_stream) >> 2)

                # ... and register the position and shape of our descriptor in memory.
                m.d.sync += [
                    descriptor_data_base_address  .eq(rom_element_pointer),
                    descriptor_length             .eq(rom_element_count),
                ]

                # Our current position may point out of bounds in case our descriptor length is a multiple
                # of the maximum packet size. We must send a ZLP now so the host knows the previous
                # packet was the end of the descriptor.
                with m.If(rom_element_count == 0):
                    m.d.comb += self.stall.eq(1)
                    m.next = 'IDLE'
                with m.Elif(position_in_stream >= rom_element_count):
                    m.next = 'SEND_ZLP'
                with m.Else():
                    m.next = 'SEND_DESCRIPTOR'


            # SEND_DESCRIPTOR -- we finally are actively streaming our descriptor; which we'll complete until
            # our descriptor is fully sent.
            with m.State('SEND_DESCRIPTOR'):
                word_in_stream = position_in_stream >> 2
                byte_in_stream = position_in_stream.bit_select(0, 2)

                m.d.comb += [
                    self.tx.valid       .eq(1),

                    # Always drive the stream from our current memory output...
                    rom_read_port.addr  .eq(descriptor_data_base_address + word_in_stream),
                    self.tx.payload     .eq(rom_read_port.data.word_select(~byte_in_stream, 8)),

                    # ... and base First and Last based on our current position in the stream.
                    self.tx.first       .eq(on_first_packet),
                    self.tx.last        .eq(on_last_packet)
                ]

                # Once a given word is accepted, we're ready to move on.
                with m.If(self.tx.ready):

                    # If we're not yet done, move to the next byte in the stream.
                    with m.If(~on_last_packet):
                        m.d.sync += [
                            position_in_stream  .eq(position_in_stream + 1),
                            bytes_sent          .eq(bytes_sent + 1),
                        ]
                        m.d.comb += rom_read_port.addr.eq(descriptor_data_base_address+(position_in_stream + 1).bit_select(2, position_in_stream.width - 2)),

                    # Otherwise, we've finished! Return to IDLE.
                    with m.Else():
                        # Reset some values, might not be really required
                        m.d.sync += [
                            descriptor_length             .eq(0),
                            descriptor_data_base_address  .eq(0)
                        ]
                        m.next = 'IDLE'

            # SEND_ZLP -- we've had an empty descriptor request, or a request that ended on a packet boundary.
            # Send a zero-length packet to end the transaction.
            with m.State('SEND_ZLP'):
                m.d.comb += [
                    # Pulse `last` without `first` to indicate a ZLP.
                    self.tx.valid.eq(1),
                    self.tx.last .eq(1),
                ]
                m.next = 'IDLE'


        # Convert our sync domain to the domain requested by the user, if necessary.
        if self._domain != "sync":
            m = DomainRenamer({"sync": self._domain})(m)

        return m


# amaranth: UnusedElaboratable=no
#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause
""" USB3 Physical-Layer modules """

from .layer import USB3PhysicalLayer

__all__ = ['USB3PhysicalLayer']



#
# This file is part of LUNA.
#
# Copyright (c) 2020-2023 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

import os
import sys
import logging

import importlib
import importlib.util

from typing import Optional

from amaranth import Record

from .core import NullPin, LUNAPlatform
from .toolchain import configure_toolchain


def _get_platform_from_string(platform):
    """ Attempts to get the most appropriate platform given a <module>:<class> specification."""

    # Attempt to split the platform into a module / name.
    module, _, name = platform.partition(':')
    if (not module) or (not name):
        raise TypeError("LUNA_PLATFORM must be in <module path>:<class name> format.")


    # If we have a filename, load the module from our file.
    module_path = os.path.expanduser(module)
    if os.path.isfile(module_path):

        # Get a reference to the platform module to be loaded...
        import_path     = "luna.gateware.platform.dynamic"
        spec            = importlib.util.spec_from_file_location(import_path, module_path)
        platform_module = importlib.util.module_from_spec(spec)

        # ... and pull in its code .
        spec.loader.exec_module(platform_module)


    # Otherwise, try to parse it as a module path.
    else:
        platform_module = importlib.import_module(module)

    # Once we have the relevant module, extract our class from it.
    platform_class = getattr(platform_module, name)
    return platform_class()


def get_appropriate_platform() -> LUNAPlatform:
    """ Attempts to return the most appropriate platform for the local configuration. """

    # If we have a LUNA_PLATFORM variable, use it.
    if os.getenv("LUNA_PLATFORM"):
        return _get_platform_from_string(os.getenv("LUNA_PLATFORM"))

    # Otherwise, try to detect an Apollo-based platform.
    platform_string, fpga_device = get_apollo_platform()
    if platform_string is None:
        raise RuntimeError(
            "Unable to autodetect a supported platform. "
            "The LUNA_PLATFORM environment variable must be set.")
    platform = _get_platform_from_string(platform_string)

    # If possible, override the platform's device type with the detected FPGA.
    if fpga_device is not None:
        platform.device = fpga_device

    return platform


def get_apollo_platform() -> (Optional[str], Optional[str]):
    """ Attempts to return a platform string for a connected Apollo-based device. """

    # Try to import Apollo and look for a debug interface.
    try:
        import apollo_fpga
        debugger = apollo_fpga.ApolloDebugger()
    except (ImportError, apollo_fpga.DebuggerNotFound):
        return None, None

    # Retrieve the version of the attached device.
    version = debugger.detect_connected_version()

    # Try to auto-detect the attached FPGA device. This can be
    # used to override the default platform's device type.
    try:
        fpga_type = debugger.get_fpga_type()
    except IOError:
        fpga_type = None

    # Check if relevant modules provide a platform string.
    try:
        import cynthion.gateware
        return cynthion.gateware.APOLLO_PLATFORMS[version], fpga_type
    except (ImportError, AttributeError, KeyError):
        pass
    try:
        import luna_boards
        return luna_boards.APOLLO_PLATFORMS[version], fpga_type
    except (ImportError, AttributeError, KeyError):
        pass

    # If none of the above modules produced a match, no platform is known.
    return None, None


class NullPin(Record):
    """ Stand-in for a I/O record. """

    def __init__(self, size=1):
        super().__init__([
            ('i', size),
            ('o', size),
            ('oe', 1),
        ])
